package com.winhc.bigdata.spark.utils

import org.apache.commons.lang3.StringUtils
import org.apache.spark.internal.Logging
import org.apache.spark.sql.types.{DataType, StructField, StructType}
import org.apache.spark.sql.{DataFrame, SparkSession}

import scala.annotation.meta.getter
import scala.collection.immutable.ListMap

/**
 * π
 */
trait LoggingUtils extends Logging {
  protected var sqlNo = 1

  @(transient@getter) protected val spark: SparkSession

  def sql(sqlString: String): DataFrame = {
    logInfo(
      s"""
         |- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
         |Job[${this.getClass.getSimpleName}].SQL[No$sqlNo.]
         |
         |$sqlString
         |- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
       """.stripMargin
    )
    println(
      s"""
         |- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
         |Job[${this.getClass.getSimpleName}].SQL[No$sqlNo.]
         |
         |$sqlString
         |- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
         |""".stripMargin)
    sqlNo += 1
    spark.sql(sqlString)
  }

  def showString(dataset: DataFrame, _numRows: Int = 20, truncate: Int = -1): String = {
    val numRows = _numRows.max(0)
    val takeResult = dataset.take(numRows + 1)
    val hasMoreData = takeResult.length > numRows
    val data = takeResult.take(numRows)

    // For array values, replace Seq and Array with square brackets
    // For cells that are beyond `truncate` characters, replace it with the
    // first `truncate-3` and "..."
    val rows: Seq[Seq[String]] = dataset.schema.fieldNames.toSeq +: data.map { row =>
      row.toSeq.map { cell =>
        val str = cell match {
          case null => "null"
          case binary: Array[Byte] => binary.map("%02X".format(_)).mkString("[", " ", "]")
          case array: Array[_] => array.mkString("[", ", ", "]")
          case seq: Seq[_] => seq.mkString("[", ", ", "]")
          case _ => cell.toString
        }
        if (truncate > 0 && str.length > truncate) {
          // do not show ellipses for strings shorter than 4 characters.
          if (truncate < 4) str.substring(0, truncate)
          else str.substring(0, truncate - 3) + "..."
        } else {
          str
        }
      }: Seq[String]
    }

    val sb = new StringBuilder
    val numCols = dataset.schema.fieldNames.length

    // Initialise the width of each column to a minimum value of '3'
    val colWidths = Array.fill(numCols)(3)

    // Compute the width of each column
    for (row <- rows) {
      for ((cell, i) <- row.zipWithIndex) {
        colWidths(i) = math.max(colWidths(i), cell.length)
      }
    }

    // Create SeparateLine
    val sep: String = colWidths.map("-" * _).addString(sb, "+", "+", "+\n").toString()

    // column names
    rows.head.zipWithIndex.map { case (cell, i) =>
      if (truncate > 0) {
        StringUtils.leftPad(cell, colWidths(i))
      } else {
        StringUtils.rightPad(cell, colWidths(i))
      }
    }.addString(sb, "|", "|", "|\n")

    sb.append(sep)

    // data
    rows.tail.map {
      _.zipWithIndex.map { case (cell, i) =>
        if (truncate > 0) {
          StringUtils.leftPad(cell.toString, colWidths(i))
        } else {
          StringUtils.rightPad(cell.toString, colWidths(i))
        }
      }.addString(sb, "|", "|", "|\n")
    }

    sb.append(sep)

    // For Data that has more than "numRows" records
    if (hasMoreData) {
      val rowsString = if (numRows == 1) "row" else "rows"
      sb.append(s"only showing top $numRows $rowsString\n")
    }

    sb.toString()
  }


  def getPartitions(t: String): Seq[String] = {
    val sql_s = s"show partitions " + t
    sql(sql_s).collect.toList.map(r => r.getString(0)).map(r => {
      if (r.contains("/")) {
        r.split("/").find(s => s.contains("ds")).map(s => s.split("=")(1)).orNull
      } else {
        r.split("=")(1)
      }
    })
  }

  def getSecondLastPartitionOrElse(t: String, default: String): String = {
    val ps = getPartitions(t)
    if (ps.length >= 2) {
      ps(ps.length - 2)
    } else {
      default
    }
  }

  def getLastPartitionsOrElse(t: String, default: String): String = {
    val ps = getPartitions(t)
    if (ps.nonEmpty) {
      ps.last
    } else {
      default
    }
  }

  def getHeadPartitionsOrElse(t: String, default: String): String = {
    val ps = getPartitions(t)
    if (ps.nonEmpty) {
      ps.head
    } else {
      default
    }
  }

  def getColumns(t: String): Seq[String] = {
    spark.table(t).columns.seq
  }

  def getSchema(map: ListMap[String, DataType]): StructType = {
    StructType(map.map(e => StructField(e._1, e._2)).toArray)
  }

  def getEnvProjectName(env: String, projectName: String): String = {
    if (env.equals("dev")) {
      projectName
    } else {
      projectName.substring(0, projectName.length - 4)
    }
  }
}
